import Q from 'q';
import React, {
    Component,
    PropTypes
} from 'react';
import {
    each,
    every,
    isUndefined,
    mapValues,
    omit,
    pick,
    uniqueId,
    values,
    isArray,
    isFunction,
    isString,
    isPlainObject,
    intersection,
    last,
    noop
} from 'lodash';
import {
    compose,
    mapProps,
    getDisplayName,
    setDisplayName,
    wrapDisplayName,
    hoistStatics
} from 'recompose';
import { resolve } from 'react-resolver';
import { connectToStores } from 'fluxible-addons-react';
import assert from 'assert';

const fluxibleContextTypes = {
    executeAction: PropTypes.func.isRequired,
    getStore: PropTypes.func.isRequired
};

const formatList = list => list.map(e => `"${e}"`).join(', ');

const assertValidResolverKeys = (displayName, resolvers) => {
    assert(isPlainObject(resolvers), `${displayName} - resolvers must be an object not a ${typeof resolvers}`);
    every(resolvers, (resolver, name) => {
        assert(isPlainObject(resolver), `${displayName} - resolver "${name}" must be an object`);
    });
    every(resolvers, ({ dependencies, isResolved, resolver, getStateFromStores, stores = [], ...extraneous }, name) => {
        const extraneousKeys = Object.keys(extraneous);
        assert.equal(extraneousKeys.length, 0, `${displayName} - resolver "${name}" unknown option(s) ${formatList(extraneousKeys)}`);
        assert(isFunction(getStateFromStores), `${displayName} - getStateFromStores must be a function`);
        assert(isArray(stores), `${displayName} - stores must be an array`);
        assert(isFunction(resolver), `${displayName} - resolver must be a function`);
        if (!isUndefined(dependencies)) {
            assert(isArray(dependencies), `${displayName} - dependencies must be an array`);
            assert(every(dependencies, isString), `${displayName} - dependencies must be strings`);
        }
        if (!isUndefined(isResolved)) {
            assert(isFunction(isResolved), `${displayName} - isResolved must be a function`);
        }
    });
};

const assertValidResolverDependencies = (displayName, resolvers) => (
    every(resolvers, ({ dependencies = [] }, propName) => (
        every(dependencies, dependency => (
            assert(resolvers.hasOwnProperty(dependency), `${displayName} - "${propName}" depends on "${dependency}", which does not exist`)
        ))
    ))
);

const assertNoResolversPropTypesCollisions = (displayName, resolvers, propTypes) => {
    const overlap = intersection(Object.keys(resolvers), Object.keys(propTypes));
    assert(overlap.length === 0, `${displayName} - cannot accept propTypes ${formatList(overlap)} which also have resolvers`);
};

const checkIsResolved = (resolvers, name, props) => {
    const {
        dependencies = [],
        isResolved
    } = resolvers[name];
    return every(dependencies, dependency => checkIsResolved(resolvers, dependency, props))
        && (isResolved ? isResolved(props) : !isUndefined(props[name]));
};

const assertResolversNotCyclic = (displayName, resolvers) => {
    const detect = path => {
        const { dependencies = [] } = resolvers[last(path)];
        dependencies.forEach(dependency => {
            const cyclic = path.includes(dependency);
            assert(!cyclic, `${displayName} - circular dependency detected ${formatList([...path.slice(path.indexOf(dependency)), dependency])}`);
        });
        dependencies.forEach(dependency => detect([...path, dependency]));
    };
    Object.keys(resolvers).forEach(r => detect([r]));
};

const buildInputPropResolver = name => ({
    dependencies: [],
    getStateFromStores: (context, props) => props[name],
    isResolved: () => true,
    resolver: noop,
    stores: []
});

export default (explicitResolvers, propTypes = {}) => hoistStatics(C => {
    const displayName = wrapDisplayName(C, 'resolver');

    assertValidResolverKeys(displayName, explicitResolvers);
    assertNoResolversPropTypesCollisions(displayName, explicitResolvers, propTypes);

    const inputPropResolvers = Object.keys(propTypes).reduce((prev, next) => ({
        ...prev,
        [next]: buildInputPropResolver(next)
    }), {});
    const resolvers = {
        ...explicitResolvers,
        ...inputPropResolvers
    };

    assertValidResolverDependencies(displayName, resolvers);
    assertResolversNotCyclic(displayName, resolvers);

    const stores = Object.keys(resolvers).reduce((prev, next) => [...prev, ...(resolvers[next].stores || [])], []);
    const safeGetStateFromStores = ({ getStore, executeAction }, props) => {
        const result = Object.keys(resolvers).reduce((prev, next) => {
            prev[next] = resolvers[next].getStateFromStores({
                getStore: store => {
                    const storeName = store.storeName || store;
                    // did we forget to add this store to the list of stores passed in?
                    // handle both strings/store objects in both the getStore calls, and stores array
                    assert(
                        stores.find(s => (s.storeName || s) === storeName),
                        `${displayName} - getStateFromStores should not call getStore(${storeName}) without adding ${storeName} to the stores array`
                    );
                    return getStore(store);
                },
                executeAction
            }, pick(props, Object.keys(propTypes)));
            return prev;
        }, {});
        assert(isPlainObject(result), `${displayName} - getStateFromStores should return an object rather than ${JSON.stringify(result)}`);
        return result;
    };

    return compose(
        setDisplayName(displayName),
        connectToStores(stores, safeGetStateFromStores, {
            executeAction: PropTypes.func.isRequired
        })
    )(class extends Component {
        static contextTypes = fluxibleContextTypes;
        static propTypes = propTypes;
        render() {
            if (every(resolvers, (resolver, propName) => checkIsResolved(resolvers, propName, this.props))) {
                return <C {...this.props}/>;
            }

            const key = uniqueId('fluxibleResolver');
            const fluxibleResolver = resolve(key, () => {
                const initialProps = safeGetStateFromStores(this.context, this.props);
                if (every(resolvers, (resolver, propName) => checkIsResolved(resolvers, propName, initialProps))) {
                    return initialProps;
                }

                const resolverDefers = mapValues(resolvers, () => Q.defer());
                each(resolverDefers, (resolverDefer, propName) => {
                    const {
                        dependencies = [],
                        resolver
                    } = resolvers[propName];

                    resolverDefer.resolve(
                        Q.all(dependencies.map(dependency => resolverDefers[dependency].promise))
                        .then(() => safeGetStateFromStores(this.context, this.props))
                        .then(props => {
                            if (checkIsResolved(resolvers, propName, props)) {
                                return Q.resolve();
                            }
                            const defer = Q.defer();
                            this.context.executeAction(actionContext => {
                                defer.resolve(actionContext.executeAction(resolver, props));
                            });
                            return defer.promise.tap(res => {
                                const state = safeGetStateFromStores(this.context, this.props);
                                assert(
                                    checkIsResolved(resolvers, propName, state),
                                    `${displayName} - unfulfilled resolution for "${propName}" in ${getDisplayName(C)} - resolver returned ${JSON.stringify(res)}, getStateFromStores returned ${JSON.stringify(state[propName])}`
                                );
                            });
                        })
                    );
                });
                return Q.all(
                    values(resolverDefers).map(({ promise }) => promise)
                ).then(() => safeGetStateFromStores(this.context, this.props));
            });
            const Container = compose(
                fluxibleResolver,
                mapProps(props => omit(props, key))
            )(C);
            return <Container {...this.props} />;
        }
    });
});
